/* $Id: parse.c 13 2023-02-22 14:28:47Z nishi $ */

#include <ncron/parse.h>

#include <ncron/util.h>
#include <ncron.h>

#include <stdlib.h>
#include <stdint.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#define CLEAN_ENTRY	free(entry);\
			mqueue_free();\
			goto skip;

void process_entry(
			char** pstart,
			char** pend,
			char** pstep,
			int min, int max, char* str, int incr
){
	*pstart = malloc(1);
	*pstart[0] = -1;
	*pend = malloc(1);
	*pend[0] = -1;
	*pstep = malloc(1);
	*pstep[0] = -1;
	char* buffer = malloc(1);
	buffer[0] = 0;
	int i = 0;
	for(i = 0; i <= sstrlen(str); i++){
		char ch = str[i];
		if(ch == ',' || ch == 0){
			int64_t c;
			char* range = buffer;
			char* step = NULL;
			if((c = sindexof(buffer, "/")) != -1){
				buffer[c] = 0;
				step = buffer + c + 1;
			}
			char start = -2;
			char end = -2;
			if((c = sindexof(range, "-")) != -1){
				range[c] = 0;
				char* strstart = range;
				char* strend = range + c + 1;
				start = atoi(strstart) + incr;
				end = atoi(strend) + incr;
			}else{
				if(strcmp(range, "*") == 0){
					start = min;
					end = max;
				}else{
					start = atoi(range) + incr;
				}
			}
			*pstart = scappend(*pstart, start);
			*pend = scappend(*pend, end);
			if(step == NULL){
				*pstep = scappend(*pstep, 1);
			}else{
				*pstep = scappend(*pstep, atoi(step));
			}
			free(buffer);
			buffer = malloc(1);
			buffer[0] = 0;
			if(ch == 0) break;
		}else{
			char* tmp = buffer;
			char* tmp2 = malloc(2);
			tmp2[0] = ch;
			tmp2[1] = 0;
			buffer = sstrcat(tmp, tmp2);
			free(tmp);
			free(tmp2);
		}
	}
	free(buffer);
}

struct ncrontab* ncron_parse_file(const char* cronpath, struct ncrontab* result, uid_t defuid, gid_t defgid, char allowchange, char isuser){
	int i = 0;
	char* buffer = malloc(1);
	buffer[0] = 0;
	int counter = 0;
	FILE* f = fopen(cronpath, "r");
	if(f != NULL){
		char ch;
		char* shell = sstrdup("/bin/sh");
		char* path = sstrdup("/bin:/sbin:/usr/bin:/usr/sbin");
		while(1){
			fread(&ch, 1, 1, f);
			if(feof(f) || ch == '\n'){
				if(counter != 0){
					if(buffer[0] != '#'){
						if(sstartswith(buffer, "SHELL=")){
							free(shell);
							shell = sstrdup(buffer + sstrlen("SHELL="));
						}else if(sstartswith(buffer, "PATH=")){
							free(path);
							path = sstrdup(buffer + sstrlen("PATH="));
						}else if(sstartswith(buffer, "MAILTO=")){
						}else if(sstartswith(buffer, "CRON_WITHIN=")){
						}else if(sstartswith(buffer, "HOME=")){
						}else{
							struct ncron_entry* entry = malloc(sizeof(*entry));
							entry->e_shell = sstrdup(shell);
							entry->e_path = sstrdup(path);
							entry->e_minute_start = NULL;
							entry->e_hour_start = NULL;
							entry->e_date_start = NULL;
							entry->e_month_start = NULL;
							entry->e_day_start = NULL;

							entry->e_minute_end = NULL;
							entry->e_hour_end = NULL;
							entry->e_date_end = NULL;
							entry->e_month_end = NULL;
							entry->e_day_end = NULL;

							entry->e_minute_step = scarray(1);
							entry->e_hour_step = scarray(1);
							entry->e_date_step = scarray(1);
							entry->e_month_step = scarray(1);
							entry->e_day_step = scarray(1);

							entry->e_reboot = 0;
							entry->e_uid = defuid;
							entry->e_gid = defgid;

							entry->e_crontab = sstrdup(cronpath);

							int64_t minute_index = sindexof(buffer, "\t ");
							if(minute_index == -1){
								CLEAN_ENTRY;
							}
							char* minute = mqueue(malloc(minute_index + 1));
							minute[minute_index] = 0;
							smemcpy(minute, buffer, minute_index);
							if(buffer[0] == '@'){
								if(strcmp(minute, "@yearly") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(-2);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(-2);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(-2);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(-2);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(6);
								}else if(strcmp(minute, "@monthly") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(-2);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(-2);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(-2);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(11);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(6);
								}else if(strcmp(minute, "@weekly") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(-2);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(-2);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(31);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(11);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(-2);
								}else if(strcmp(minute, "@daily") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(-2);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(-2);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(31);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(11);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(6);
								}else if(strcmp(minute, "@hourly") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(-2);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(23);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(31);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(11);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(6);
								}else if(strcmp(minute, "@minutely") == 0){
									entry->e_minute_start = scarray(0);
									entry->e_minute_end = scarray(59);

									entry->e_hour_start = scarray(0);
									entry->e_hour_end = scarray(23);

									entry->e_date_start = scarray(1);
									entry->e_date_end = scarray(31);

									entry->e_month_start = scarray(0);
									entry->e_month_end = scarray(11);

									entry->e_day_start = scarray(0);
									entry->e_day_end = scarray(6);
								}else if(strcmp(minute, "@reboot") == 0){
									entry->e_reboot = 1;
								}else{
									CLEAN_ENTRY;
								}
								char* user;
								int64_t user_index = sindexof(buffer + minute_index + scontinue(buffer + minute_index, "\t "), "\t ");
								if(isuser == 0){
									if(user_index != -1){
										user = mqueue(malloc(user_index + 1));
										user[user_index] = 0;
										smemcpy(user, minute_index + buffer + scontinue(buffer + minute_index, "\t "), user_index);
									}
								}
								entry->e_command = NULL;
								if(user != NULL && isuser == 0){
									entry->e_command = sstrdup(user_index + scontinue(buffer + minute_index, "\t ") + minute_index + buffer + 1);
									if(user != NULL && allowchange == 1){
										struct passwd* pwd = getpwnam(user);
										if(pwd != NULL){
											entry->e_uid = pwd->pw_uid;
											entry->e_gid = pwd->pw_gid;
										}
									}
								}else{
									entry->e_command = sstrdup(scontinue(buffer + minute_index, "\t ") + minute_index + buffer);
								}
							}else{
								int spaces = scontinue(buffer + minute_index, "\t ");
								int64_t hour_index = sindexof(buffer + minute_index + spaces, "\t ");
								if(hour_index == -1){
									CLEAN_ENTRY;
								}
								char* hour = mqueue(malloc(hour_index + 1));
								hour[hour_index] = 0;
								smemcpy(hour, minute_index + buffer + spaces, hour_index);
	
								spaces += scontinue(buffer + minute_index + spaces + hour_index, "\t ");
								int64_t date_index = sindexof(buffer + spaces + hour_index + minute_index, "\t ");
								if(date_index == -1){
									CLEAN_ENTRY;
								}
								char* date = mqueue(malloc(date_index + 1));
								date[date_index] = 0;
								smemcpy(date, spaces + hour_index + minute_index + buffer, date_index);

								spaces += scontinue(buffer + spaces + minute_index + hour_index + date_index, "\t ");
								int64_t month_index = sindexof(buffer + spaces + date_index + hour_index + minute_index, "\t ");
								if(month_index == -1){
									CLEAN_ENTRY;
								}
								char* month = mqueue(malloc(month_index + 1));
								month[month_index] = 0;
								smemcpy(month, spaces + date_index + hour_index + minute_index + buffer, month_index);

								spaces += scontinue(buffer + spaces + minute_index + hour_index + date_index + month_index, "\t ");
								int64_t day_index = sindexof(buffer + spaces + month_index + date_index + hour_index + minute_index, "\t ");
								if(day_index == -1){
									CLEAN_ENTRY;
								}
								char* day = mqueue(malloc(day_index + 1));
								day[day_index] = 0;
								smemcpy(day, spaces + month_index + date_index + hour_index + minute_index + buffer, day_index);

								spaces += scontinue(buffer + spaces + minute_index + hour_index + date_index + day_index + month_index, "\t ");
								int64_t user_index = sindexof(buffer + spaces + day_index + month_index + date_index + hour_index + minute_index, "\t ");
								char* user = NULL;
								char* cmd = NULL;
								if(user_index == -1){
									cmd = buffer + day_index + month_index + date_index + hour_index + minute_index + spaces;
									user = NULL;
								}else{
									user = mqueue(malloc(user_index + 1));
									user[user_index] = 0;
									smemcpy(user, day_index + spaces + month_index + date_index + hour_index + minute_index + buffer, user_index);
								}

								process_entry(
										&entry->e_minute_start,
										&entry->e_minute_end,
										&entry->e_minute_step,
										0, 59, minute, 0
								);
								process_entry(
										&entry->e_hour_start,
										&entry->e_hour_end,
										&entry->e_hour_step,
										0, 23, hour, 0
								);
								process_entry(
										&entry->e_date_start,
										&entry->e_date_end,
										&entry->e_date_step,
										1, 31, date, 0
								);
								process_entry(
										&entry->e_month_start,
										&entry->e_month_end,
										&entry->e_month_step,
										0, 11, month, -1
								);
								process_entry(
										&entry->e_day_start,
										&entry->e_day_end,
										&entry->e_day_step,
										0, 6, day, 0
								);

								int orig_spaces = spaces;
								if(isuser == 0){
									spaces += scontinue(buffer + spaces + minute_index + hour_index + date_index + day_index + month_index + user_index, "\t ");
									entry->e_command = sstrdup(user_index + day_index + month_index + date_index + hour_index + minute_index + buffer + spaces);
									if(allowchange == 1 && user != NULL){
										struct passwd* pwd = getpwnam(user);
										if(pwd != NULL){
											entry->e_uid = pwd->pw_uid;
											entry->e_gid = pwd->pw_gid;
										}
									}
								}else{
									entry->e_command = sstrdup(day_index + month_index + date_index + hour_index + minute_index + buffer + spaces);
								}
							}
							mqueue_free();

							result->tabs = realloc(result->tabs, sizeof(*result->tabs) * (result->tab_counter + 2));
							result->tabs[result->tab_counter] = entry;
							result->tabs[result->tab_counter + 1] = NULL;
							result->tab_counter++;
						}
					}
				}
skip:
				free(buffer);
				buffer = malloc(1);
				buffer[0] = 0;
				counter = 0;
				if(feof(f)) break;
			}else{
				buffer = realloc(buffer, counter + 2);
				buffer[counter] = ch;
				buffer[counter + 1] = 0;
				counter++;
			}
		}
		fclose(f);
	}
	return result;
}

struct ncrontab* ncron_parse(){
	struct stat s;
	struct ncrontab* result = malloc(sizeof(*result));
	result->crontab_count = 0;
	result->crontabs = malloc(sizeof(*result->crontabs));
	result->crontabs[0] = NULL;
	result->tabs = malloc(sizeof(*result->tabs));
	result->tabs[0] = NULL;
	result->crontab_mtimes = malloc(0);
	result->uids = malloc(0);
	result->gids = malloc(0);
	result->isusers = malloc(0);
	result->tab_counter = 0;
	result = ncron_parse_file("/etc/crontab", result, 0, 0, 1, 0);
	result->crontabs = realloc(result->crontabs, sizeof(*result->crontabs) * (2 + result->crontab_count));
	result->crontabs[result->crontab_count] = sstrdup("/etc/crontab");
	result->crontabs[result->crontab_count + 1] = NULL;
	result->crontab_mtimes = realloc(result->crontab_mtimes, sizeof(*result->crontab_mtimes) * (1 + result->crontab_count));
	result->uids = realloc(result->uids, sizeof(*result->uids) * (1 + result->crontab_count));
	result->gids = realloc(result->gids, sizeof(*result->gids) * (1 + result->crontab_count));
	result->isusers = realloc(result->isusers, sizeof(*result->isusers) * (1 + result->crontab_count));
	stat("/etc/crontab", &s);
	result->crontab_mtimes[result->crontab_count] = s.st_mtime;
	result->uids[result->crontab_count] = 0;
	result->gids[result->crontab_count] = 0;
	result->isusers[result->crontab_count] = 0;
	result->crontab_count++;
	DIR* dir;
	struct dirent* d;
      	dir = opendir("/etc/cron.d");
	if(dir != NULL){
		while((d = readdir(dir)) != NULL){
			if(strcmp(d->d_name, "..") == 0 || strcmp(d->d_name, ".") == 0) continue;
			char* path = sstrcat("/etc/cron.d/", d->d_name);
			result = ncron_parse_file(path, result, 0, 0, 1, 0);

			result->crontabs = realloc(result->crontabs, sizeof(*result->crontabs) * (2 + result->crontab_count));
			result->crontabs[result->crontab_count] = sstrcat("/etc/cron.d/", d->d_name);
			result->crontabs[result->crontab_count + 1] = NULL;
			result->crontab_mtimes = realloc(result->crontab_mtimes, sizeof(*result->crontab_mtimes) * (1 + result->crontab_count));
			result->uids = realloc(result->uids, sizeof(*result->uids) * (1 + result->crontab_count));
			result->gids = realloc(result->gids, sizeof(*result->gids) * (1 + result->crontab_count));
			result->isusers = realloc(result->isusers, sizeof(*result->isusers) * (1 + result->crontab_count));
			stat(path, &s);
			result->crontab_mtimes[result->crontab_count] = s.st_mtime;
			result->uids[result->crontab_count] = 0;
			result->gids[result->crontab_count] = 0;
			result->isusers[result->crontab_count] = 0;
			result->crontab_count++;
			free(path);
		}
		closedir(dir);
	}
      	dir = opendir("/var/cron/tabs");
	if(dir != NULL){
		while((d = readdir(dir)) != NULL){
			if(strcmp(d->d_name, "..") == 0 || strcmp(d->d_name, ".") == 0) continue;
			uid_t uid;
			gid_t gid;
			struct passwd* pwd = getpwnam(d->d_name);
			if(pwd != NULL){
				uid = pwd->pw_uid;
				gid = pwd->pw_gid;
				char* path = sstrcat("/var/cron/tabs/", d->d_name);
				result = ncron_parse_file(path, result, uid, gid, 0, 1);

				result->crontabs = realloc(result->crontabs, sizeof(*result->crontabs) * (2 + result->crontab_count));
				result->crontabs[result->crontab_count] = sstrcat("/var/cron/tabs/", d->d_name);
				result->crontabs[result->crontab_count + 1] = NULL;
				result->crontab_mtimes = realloc(result->crontab_mtimes, sizeof(*result->crontab_mtimes) * (1 + result->crontab_count));
				result->uids = realloc(result->uids, sizeof(*result->uids) * (1 + result->crontab_count));
				result->gids = realloc(result->gids, sizeof(*result->gids) * (1 + result->crontab_count));
				result->isusers = realloc(result->isusers, sizeof(*result->isusers) * (1 + result->crontab_count));
				stat(path, &s);
				result->crontab_mtimes[result->crontab_count] = s.st_mtime;
				result->uids[result->crontab_count] = uid;
				result->gids[result->crontab_count] = gid;
				result->isusers[result->crontab_count] = 1;
				result->crontab_count++;
				free(path);
			}
		}
		closedir(dir);
	}
	return result;
}

void ncron_entry_free(struct ncron_entry* entry){
	if(entry->e_minute_start != NULL) free(entry->e_minute_start);
	if(entry->e_hour_start != NULL) free(entry->e_hour_start);
	if(entry->e_date_start != NULL) free(entry->e_date_start);
	if(entry->e_month_start != NULL) free(entry->e_month_start);
	if(entry->e_day_start != NULL) free(entry->e_day_start);

	if(entry->e_minute_end != NULL) free(entry->e_minute_end);
	if(entry->e_hour_end != NULL) free(entry->e_hour_end);
	if(entry->e_date_end != NULL) free(entry->e_date_end);
	if(entry->e_month_end != NULL) free(entry->e_month_end);
	if(entry->e_day_end != NULL) free(entry->e_day_end);

	if(entry->e_minute_step != NULL) free(entry->e_minute_step);
	if(entry->e_hour_step != NULL) free(entry->e_hour_step);
	if(entry->e_date_step != NULL) free(entry->e_date_step);
	if(entry->e_month_step != NULL) free(entry->e_month_step);
	if(entry->e_day_step != NULL) free(entry->e_day_step);

	free(entry->e_shell);
	free(entry->e_path);
	free(entry->e_command);
	free(entry->e_crontab);
	free(entry);
}

struct ncrontab* ncron_check_mtime(struct ncrontab* tab){
	int i;
	struct stat s;
	char** update_crons = malloc(sizeof(*update_crons));
	update_crons[0] = NULL;
	uid_t* uids = malloc(0);
	gid_t* gids = malloc(0);
	char* isusers = malloc(0);
	time_t* mtimes = malloc(0);
	int cron_count = 0;
	char** exists = malloc(sizeof(*exists));
	exists[0] = NULL;
	int* exists_index = malloc(0);
	int exist_count = 0;
	for(i = 0; tab->crontabs[i] != NULL; i++){
		if(0 == stat(tab->crontabs[i], &s)){
			exists = realloc(exists, sizeof(*exists) * (exist_count + 2));
			exists[exist_count] = tab->crontabs[i];
			exists_index[exist_count] = i;
			exists[exist_count + 1] = NULL;

			uids = realloc(uids, sizeof(*uids) * (exist_count + 1));
			uids[exist_count] = tab->uids[i];

			gids = realloc(gids, sizeof(*gids) * (exist_count + 1));
			gids[exist_count] = tab->gids[i];

			isusers = realloc(isusers, sizeof(*isusers) * (exist_count + 1));
			isusers[exist_count] = tab->isusers[i];

			mtimes = realloc(mtimes, sizeof(*mtimes) * (exist_count + 1));
			mtimes[exist_count] = s.st_mtime;

			exist_count++;
		}else{
			continue;
		}
		if(s.st_mtime != tab->crontab_mtimes[i]){
			update_crons = realloc(update_crons, sizeof(*update_crons) * (cron_count + 2));
			update_crons[cron_count] = sstrdup(tab->crontabs[i]);
			update_crons[cron_count + 1] = NULL;

			cron_count++;
		}
	}
	char** crontabs = malloc(sizeof(*crontabs));
	crontabs[0] = NULL;
	tab->crontab_count = 0;
	for(i = 0; exists[i] != NULL; i++){
		crontabs = realloc(crontabs, sizeof(*crontabs) * (tab->crontab_count + 2));
		crontabs[tab->crontab_count] = sstrdup(exists[i]);
		crontabs[tab->crontab_count + 1] = NULL;

		uids = realloc(uids, sizeof(*uids) * (tab->crontab_count + 1));
		uids[tab->crontab_count] = tab->uids[exists_index[i]];

		gids = realloc(gids, sizeof(*gids) * (tab->crontab_count + 1));
		gids[tab->crontab_count] = tab->gids[exists_index[i]];

		isusers = realloc(isusers, sizeof(*isusers) * (tab->crontab_count + 1));
		isusers[tab->crontab_count] = tab->isusers[exists_index[i]];

		tab->crontab_count++;
	}
	for(i = 0; tab->crontabs[i] != NULL; i++){
		free(tab->crontabs[i]);
	}
	free(tab->crontabs);
	free(tab->uids);
	free(tab->gids);
	free(tab->crontab_mtimes);
	free(tab->isusers);
	tab->crontabs = crontabs;
	tab->uids = uids;
	tab->gids = gids;
	tab->isusers = isusers;
	tab->crontab_mtimes = mtimes;
	struct ncron_entry** entries = malloc(sizeof(*entries));
	entries[0] = NULL;
	cron_count = 0;
	int j;
	for(i = 0; tab->tabs[i] != NULL; i++){
		char matched = 0;
		for(j = 0; update_crons[j] != NULL; j++){
			if(strcmp(tab->tabs[i]->e_crontab, update_crons[j]) == 0){
				matched = 1;
				break;
			}
		}
		char tab_exists = 0;
		for(j = 0; exists[j] != NULL; j++){
			if(strcmp(tab->tabs[i]->e_crontab, exists[j]) == 0){
				tab_exists = 1;
				break;
			}
		}
		if(matched == 0 && tab_exists == 1){
			entries = realloc(entries, sizeof(*entries) * (cron_count + 2));
			entries[cron_count] = tab->tabs[i];
			entries[cron_count + 1] = NULL;
			cron_count++;
		}else{
			ncron_entry_free(tab->tabs[i]);
		}
	}
	free(tab->tabs);
	tab->tabs = entries;
	tab->tab_counter = cron_count;
	for(i = 0; update_crons[i] != NULL; i++){
		tab = ncron_parse_file(update_crons[i], tab, uids[i], gids[i], 1, isusers[i]);
	}
	for(i = 0; update_crons[i] != NULL; i++){
		free(update_crons[i]);
	}
	free(uids);
	free(gids);
	free(isusers);
	free(exists);
	free(update_crons);
	return tab;
}

void ncron_ncrontab_free(struct ncrontab* tab){
	int i;
	for(i = 0; tab->tabs[i] != NULL; i++){
		struct ncron_entry* entry = tab->tabs[i];
		ncron_entry_free(entry);
	}
	free(tab->tabs);
	for(i = 0; tab->crontabs[i] != NULL; i++){
		free(tab->crontabs[i]);
	}
	free(tab->crontabs);
	free(tab->crontab_mtimes);
	free(tab->uids);
	free(tab->gids);
	free(tab->isusers);
	free(tab);
}
